# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pickle

import pyomo.common.unittest as unittest

from pyomo.core.base.range import (
    NumericRange as NR,
    NonNumericRange as NNR,
    RangeProduct as RP,
    AnyRange,
    RangeDifferenceError,
)
from pyomo.core.base.set import Any

_inf = float('inf')


class TestNumericRange(unittest.TestCase):
    def test_init(self):
        a = NR(None, None, 0)
        self.assertEqual(a.start, -_inf)
        self.assertEqual(a.end, _inf)
        self.assertEqual(a.step, 0)

        a = NR(-_inf, _inf, 0)
        self.assertEqual(a.start, -_inf)
        self.assertEqual(a.end, _inf)
        self.assertEqual(a.step, 0)

        a = NR(0, None, 0)
        self.assertEqual(a.start, 0)
        self.assertEqual(a.end, _inf)
        self.assertEqual(a.step, 0)

        a = NR(0, 0, 0)
        self.assertEqual(a.start, 0)
        self.assertEqual(a.end, 0)
        self.assertEqual(a.step, 0)

        with self.assertRaisesRegex(
            ValueError, '.*start must be <= end for continuous ranges'
        ):
            NR(0, -1, 0)

        with self.assertRaisesRegex(ValueError, '.*start must not be None'):
            NR(None, None, 1)

        with self.assertRaisesRegex(ValueError, '.*step must be int'):
            NR(None, None, 1.5)

        with self.assertRaisesRegex(
            ValueError, '.*start, end ordering incompatible with step direction'
        ):
            NR(0, 1, -1)

        with self.assertRaisesRegex(
            ValueError, '.*start, end ordering incompatible with step direction'
        ):
            NR(1, 0, 1)

        with self.assertRaisesRegex(
            ValueError, r'\[0:1\] is discrete, but passed closed=\(False, True\)'
        ):
            NR(0, 1, 1, "(]")

        a = NR(0, None, 1)
        self.assertEqual(a.start, 0)
        self.assertEqual(a.end, _inf)
        self.assertEqual(a.step, 1)

        a = NR(0, None, -1)
        self.assertEqual(a.start, 0)
        self.assertEqual(a.end, -_inf)
        self.assertEqual(a.step, -1)

        a = NR(0, 5, 1)
        self.assertEqual(a.start, 0)
        self.assertEqual(a.end, 5)
        self.assertEqual(a.step, 1)

        a = NR(0, 5, 2)
        self.assertEqual(a.start, 0)
        self.assertEqual(a.end, 4)
        self.assertEqual(a.step, 2)

        a = NR(0, 5, 10)
        self.assertEqual(a.start, 0)
        self.assertEqual(a.end, 0)
        self.assertEqual(a.step, 0)

        a = NR(0, 5.5, 1)
        self.assertEqual(a.start, 0)
        self.assertEqual(a.end, 5)
        self.assertEqual(a.step, 1)

        a = NR(0.5, 5.5, 1)
        self.assertEqual(a.start, 0.5)
        self.assertEqual(a.end, 5.5)
        self.assertEqual(a.step, 1)

        with self.assertRaisesRegex(
            ValueError, '.*start, end ordering incompatible with step'
        ):
            NR(0, -1, 1)

        with self.assertRaisesRegex(
            ValueError, '.*start, end ordering incompatible with step'
        ):
            NR(0, 1, -2)

    def test_str(self):
        self.assertEqual(str(NR(1, 10, 0)), "[1..10]")
        self.assertEqual(str(NR(1, 10, 1)), "[1:10]")
        self.assertEqual(str(NR(1, 10, 3)), "[1:10:3]")
        self.assertEqual(str(NR(1, 1, 1)), "[1]")

    def test_eq(self):
        self.assertEqual(NR(1, 1, 1), NR(1, 1, 1))
        self.assertEqual(NR(1, None, 0), NR(1, None, 0))
        self.assertEqual(NR(0, 10, 3), NR(0, 9, 3))

        self.assertNotEqual(NR(1, 1, 1), NR(1, None, 1))
        self.assertNotEqual(NR(1, None, 0), NR(1, None, 1))
        self.assertNotEqual(NR(0, 10, 3), NR(0, 8, 3))

    def test_contains(self):
        # Test non-numeric values
        self.assertNotIn(None, NR(None, None, 0))
        self.assertNotIn(None, NR(0, 10, 0))
        self.assertNotIn(None, NR(0, None, 1))
        self.assertNotIn(None, NR(0, 10, 1))

        self.assertNotIn('1', NR(None, None, 0))
        self.assertNotIn('1', NR(0, 10, 0))
        self.assertNotIn('1', NR(0, None, 1))
        self.assertNotIn('1', NR(0, 10, 1))

        # Test continuous ranges
        self.assertIn(0, NR(0, 10, 0))
        self.assertIn(0, NR(None, 10, 0))
        self.assertIn(0, NR(0, None, 0))
        self.assertIn(1, NR(0, 10, 0))
        self.assertIn(1, NR(None, 10, 0))
        self.assertIn(1, NR(0, None, 0))
        self.assertIn(10, NR(0, 10, 0))
        self.assertIn(10, NR(None, 10, 0))
        self.assertIn(10, NR(0, None, 0))
        self.assertNotIn(-1, NR(0, 10, 0))
        self.assertNotIn(-1, NR(0, None, 0))
        self.assertNotIn(11, NR(0, 10, 0))
        self.assertNotIn(11, NR(None, 10, 0))

        self.assertNotIn(0, NR(0.5, 10.5, 0))
        self.assertIn(0, NR(None, 10.5, 0))
        self.assertNotIn(0, NR(0.5, None, 0))
        self.assertNotIn(11, NR(0.5, 10.5, 0))
        self.assertNotIn(11, NR(None, 10.5, 0))
        self.assertIn(11, NR(0.5, None, 0))
        self.assertIn(1.5, NR(0.5, 10.5, 0))
        self.assertIn(1.5, NR(None, 10.5, 0))
        self.assertIn(1.5, NR(0.5, None, 0))

        # test discrete ranges (both increasing & decreasing)
        self.assertIn(0, NR(0, 10, 1))
        self.assertIn(0, NR(10, None, -1))
        self.assertIn(0, NR(0, None, 1))
        self.assertIn(1, NR(0, 10, 1))
        self.assertIn(1, NR(10, None, -1))
        self.assertIn(1, NR(0, None, 1))
        self.assertIn(10, NR(0, 10, 1))
        self.assertIn(10, NR(10, None, -1))
        self.assertIn(10, NR(0, None, 1))
        self.assertNotIn(-1, NR(0, 10, 1))
        self.assertNotIn(-1, NR(0, None, 1))
        self.assertNotIn(11, NR(0, 10, 1))
        self.assertNotIn(11, NR(10, None, -1))
        self.assertNotIn(1.1, NR(0, 10, 1))
        self.assertNotIn(1.1, NR(10, None, -1))
        self.assertNotIn(1.1, NR(0, None, 1))

        self.assertNotIn(0, NR(0.5, 10.5, 1))
        self.assertNotIn(0, NR(10.5, None, -1))
        self.assertNotIn(0, NR(0.5, None, 1))
        self.assertNotIn(11, NR(0.5, 10.5, 1))
        self.assertNotIn(11, NR(10.5, None, -1))
        self.assertNotIn(11, NR(0.5, None, 1))
        self.assertIn(1.5, NR(0.5, 10.5, 1))
        self.assertIn(1.5, NR(10.5, None, -1))
        self.assertIn(1.5, NR(0.5, None, 1))

        # test discrete ranges (increasing/decreasing by 2)
        self.assertIn(0, NR(0, 10, 2))
        self.assertIn(0, NR(0, -10, -2))
        self.assertIn(0, NR(10, None, -2))
        self.assertIn(0, NR(0, None, 2))
        self.assertIn(2, NR(0, 10, 2))
        self.assertIn(-2, NR(0, -10, -2))
        self.assertIn(2, NR(10, None, -2))
        self.assertIn(2, NR(0, None, 2))
        self.assertIn(10, NR(0, 10, 2))
        self.assertIn(-10, NR(0, -10, -2))
        self.assertIn(10, NR(10, None, -2))
        self.assertIn(10, NR(0, None, 2))
        self.assertNotIn(1, NR(0, 10, 2))
        self.assertNotIn(-1, NR(0, -10, -2))
        self.assertNotIn(1, NR(10, None, -2))
        self.assertNotIn(1, NR(0, None, 2))
        self.assertNotIn(-2, NR(0, 10, 2))
        self.assertNotIn(2, NR(0, -10, -2))
        self.assertNotIn(-2, NR(0, None, 2))
        self.assertNotIn(12, NR(0, 10, 2))
        self.assertNotIn(-12, NR(0, -10, -2))
        self.assertNotIn(12, NR(10, None, -2))
        self.assertNotIn(1.1, NR(0, 10, 2))
        self.assertNotIn(1.1, NR(0, -10, -2))
        self.assertNotIn(-1.1, NR(10, None, -2))
        self.assertNotIn(1.1, NR(0, None, 2))

        # test special cases (for implicit numpy ndarray compatibility)
        self.assertNotIn('z', NR(0, None, 0))
        self.assertNotIn(['z'], NR(0, None, 0))
        self.assertIn([1], NR(0, None, 0))
        self.assertNotIn([-1], NR(0, None, 0))
        self.assertIn({0: 1}, NR(0, None, 0))
        self.assertNotIn({0: -1}, NR(0, None, 0))
        self.assertNotIn({1: 1}, NR(0, None, 0))

        class _Unrelated:
            pass

        self.assertNotIn(_Unrelated(), NR(0, None, 0))
        self.assertNotIn(_Unrelated, NR._types_comparable_to_int)

        class _NonComparable(_Unrelated):
            def __init__(self, val):
                self.val = val

            def __sub__(self, other):
                return self

            def __gt__(self, other):
                return True

            def __le__(self, other):
                return True

        self.assertNotIn(_NonComparable(1), NR(0, None, 0))
        self.assertNotIn(_NonComparable, NR._types_comparable_to_int)

        class _NotCastable(_NonComparable):
            def __lt__(self, other):
                return True

            def __eq__(self, other):
                return True

            def __ne__(self, other):
                return True

        self.assertNotIn(_NotCastable(1), NR(0, None, 0))
        self.assertNotIn(_NotCastable, NR._types_comparable_to_int)

        class _Custom:
            def __init__(self, val):
                self.val = val

            def __lt__(self, other):
                return self.val < other

            def __gt__(self, other):
                return self.val > other

            def __le__(self, other):
                return self.val <= other

            def __ge__(self, other):
                return self.val >= other

            def __eq__(self, other):
                return self.val == other

            def __sub__(self, other):
                return self.val - other

        self.assertIn(_Custom(1), NR(0, None, 0))
        self.assertIn(_Custom, NR._types_comparable_to_int)
        NR._types_comparable_to_int.discard(_Custom)

    def test_isdisjoint(self):
        def _isdisjoint(expected_result, a, b):
            self.assertIs(expected_result, a.isdisjoint(b))
            self.assertIs(expected_result, b.isdisjoint(a))

        #
        # Simple continuous ranges
        _isdisjoint(True, NR(0, 1, 0), NR(2, 3, 0))
        _isdisjoint(True, NR(2, 3, 0), NR(0, 1, 0))

        _isdisjoint(False, NR(0, 1, 0), NR(1, 2, 0))
        _isdisjoint(False, NR(0, 1, 0), NR(0.5, 2, 0))
        _isdisjoint(False, NR(0, 1, 0), NR(0, 2, 0))
        _isdisjoint(False, NR(0, 1, 0), NR(-1, 2, 0))

        _isdisjoint(False, NR(0, 1, 0), NR(-1, 0, 0))
        _isdisjoint(False, NR(0, 1, 0), NR(-1, 0.5, 0))
        _isdisjoint(False, NR(0, 1, 0), NR(-1, 1, 0))
        _isdisjoint(False, NR(0, 1, 0), NR(-1, 2, 0))

        _isdisjoint(True, NR(0, 1, 0, (True, False)), NR(1, 2, 0))
        _isdisjoint(True, NR(0, 1, 0, (False, True)), NR(-1, 0, 0))

        #
        # Continuous to discrete ranges (positive step)
        #
        _isdisjoint(True, NR(0, 1, 0), NR(2, 3, 1))
        _isdisjoint(True, NR(2, 3, 0), NR(0, 1, 1))

        _isdisjoint(False, NR(0, 1, 0), NR(-1, 2, 1))

        _isdisjoint(False, NR(0.25, 1, 0), NR(1, 2, 1))
        _isdisjoint(False, NR(0.25, 1, 0), NR(0.5, 2, 1))
        _isdisjoint(False, NR(0.25, 1, 0), NR(0, 2, 1))
        _isdisjoint(False, NR(0.25, 1, 0), NR(-1, 2, 1))

        _isdisjoint(False, NR(0, 0.75, 0), NR(-1, 0, 1))
        _isdisjoint(False, NR(0, 0.75, 0), NR(-1, 0.5, 1))
        _isdisjoint(False, NR(0, 0.75, 0), NR(-1, 1, 1))
        _isdisjoint(False, NR(0, 0.75, 0), NR(-1, 2, 1))

        _isdisjoint(True, NR(0.1, 0.9, 0), NR(-1, 0, 1))
        _isdisjoint(True, NR(0.1, 0.9, 0), NR(-1, 0.5, 1))
        _isdisjoint(True, NR(0.1, 0.9, 0), NR(-1, 1, 1))
        _isdisjoint(True, NR(0.1, 0.9, 0), NR(-1, 2, 1))

        _isdisjoint(False, NR(-0.1, 1.1, 0), NR(-1, 2, 1))
        _isdisjoint(False, NR(-0.1, 1.1, 0), NR(-2, 0, 2))
        _isdisjoint(True, NR(-0.1, 1.1, 0), NR(-1, -1, 1))
        _isdisjoint(True, NR(-0.1, 1.1, 0), NR(-2, -1, 1))

        # (additional edge cases)
        _isdisjoint(False, NR(0, 1, 0, closed=(True, True)), NR(-1, 2, 1))
        _isdisjoint(False, NR(0, 1, 0, closed=(True, False)), NR(-1, 2, 1))
        _isdisjoint(False, NR(0, 1, 0, closed=(False, True)), NR(-1, 2, 1))
        _isdisjoint(True, NR(0, 1, 0, closed=(False, False)), NR(-1, 2, 1))
        _isdisjoint(True, NR(0.1, 1, 0, closed=(True, False)), NR(-1, 2, 1))
        _isdisjoint(True, NR(0, 0.9, 0, closed=(False, True)), NR(-1, 2, 1))
        _isdisjoint(False, NR(0, 0.99, 0), NR(-1, 1, 1))
        _isdisjoint(True, NR(0.001, 0.99, 0), NR(-1, 1, 1))

        #
        # Continuous to discrete ranges (negative step)
        #
        _isdisjoint(True, NR(0, 1, 0), NR(3, 2, -1))
        _isdisjoint(True, NR(2, 3, 0), NR(1, 0, -1))

        _isdisjoint(False, NR(0, 1, 0), NR(2, -1, -1))

        _isdisjoint(False, NR(0.25, 1, 0), NR(2, 1, -1))
        _isdisjoint(False, NR(0.25, 1, 0), NR(2, 0.5, -1))
        _isdisjoint(False, NR(0.25, 1, 0), NR(2, 0, -1))
        _isdisjoint(False, NR(0.25, 1, 0), NR(2, -1, -1))

        _isdisjoint(False, NR(0, 0.75, 0), NR(0, -1, -1))
        _isdisjoint(False, NR(0, 0.75, 0), NR(0.5, -1, -1))
        _isdisjoint(False, NR(0, 0.75, 0), NR(1, -1, -1))
        _isdisjoint(False, NR(0, 0.75, 0), NR(2, -1, -1))

        # (additional edge cases)
        _isdisjoint(False, NR(0, 0.99, 0), NR(1, -1, -1))
        _isdisjoint(True, NR(0.01, 0.99, 0), NR(1, -1, -1))

        #
        # Discrete to discrete sets
        #
        _isdisjoint(False, NR(0, 10, 2), NR(2, 10, 2))
        _isdisjoint(True, NR(0, 10, 2), NR(1, 10, 2))

        _isdisjoint(False, NR(0, 50, 5), NR(0, 50, 7))
        _isdisjoint(False, NR(0, 34, 5), NR(0, 34, 7))
        _isdisjoint(False, NR(5, 50, 5), NR(7, 50, 7))
        _isdisjoint(True, NR(5, 34, 5), NR(7, 34, 7))
        _isdisjoint(False, NR(5, 50, 5), NR(49, 7, -7))
        _isdisjoint(True, NR(5, 34, 5), NR(28, 7, -7))

        _isdisjoint(True, NR(0.25, 10, 1), NR(0.5, 20, 1))
        _isdisjoint(True, NR(0.25, 10, 1), NR(0.5, 20, 2))
        _isdisjoint(True, NR(0, 100, 2), NR(1, 100, 4))
        _isdisjoint(True, NR(0, None, 2), NR(1, None, 4))
        _isdisjoint(True, NR(0.25, None, 1), NR(0.5, None, 1))

        # 1, 8, 15, 22, 29, 36
        _isdisjoint(False, NR(0, None, 5), NR(1, None, 7))
        _isdisjoint(False, NR(0, None, -5), NR(1, None, -7))
        # 2, 9, 16, 23, 30, 37
        _isdisjoint(True, NR(0, None, 5), NR(23, None, -7))
        # 0, 7, 14, 21, 28, 35
        _isdisjoint(False, NR(0, None, 5), NR(28, None, -7))

    def test_issubset(self):
        # Continuous-continuous
        self.assertTrue(NR(0, 10, 0).issubset(NR(0, 10, 0)))
        self.assertTrue(NR(1, 10, 0).issubset(NR(0, 10, 0)))
        self.assertTrue(NR(0, 9, 0).issubset(NR(0, 10, 0)))
        self.assertTrue(NR(1, 9, 0).issubset(NR(0, 10, 0)))
        self.assertFalse(NR(0, 11, 0).issubset(NR(0, 10, 0)))
        self.assertFalse(NR(-1, 10, 0).issubset(NR(0, 10, 0)))

        self.assertTrue(NR(0, 10, 0).issubset(NR(0, None, 0)))
        self.assertTrue(NR(1, 10, 0).issubset(NR(0, None, 0)))
        self.assertFalse(NR(-1, 10, 0).issubset(NR(0, None, 0)))

        self.assertTrue(NR(0, 10, 0).issubset(NR(None, 10, 0)))
        self.assertTrue(NR(0, 9, 0).issubset(NR(None, 10, 0)))
        self.assertFalse(NR(0, 11, 0).issubset(NR(None, 10, 0)))

        self.assertTrue(NR(0, None, 0).issubset(NR(None, None, 0)))
        self.assertTrue(NR(0, None, 0).issubset(NR(-1, None, 0)))
        self.assertTrue(NR(0, None, 0).issubset(NR(0, None, 0)))
        self.assertFalse(NR(0, None, 0).issubset(NR(1, None, 0)))
        self.assertFalse(NR(0, None, 0).issubset(NR(None, 1, 0)))

        self.assertTrue(NR(None, 0, 0).issubset(NR(None, None, 0)))
        self.assertTrue(NR(None, 0, 0).issubset(NR(None, 1, 0)))
        self.assertTrue(NR(None, 0, 0).issubset(NR(None, 0, 0)))
        self.assertFalse(NR(None, 0, 0).issubset(NR(None, -1, 0)))
        self.assertFalse(NR(None, 0, 0).issubset(NR(0, None, 0)))

        B = True, True
        self.assertTrue(NR(0, 1, 0, (True, True)).issubset(NR(0, 1, 0, B)))
        self.assertTrue(NR(0, 1, 0, (True, False)).issubset(NR(0, 1, 0, B)))
        self.assertTrue(NR(0, 1, 0, (False, True)).issubset(NR(0, 1, 0, B)))
        self.assertTrue(NR(0, 1, 0, (False, False)).issubset(NR(0, 1, 0, B)))

        B = True, False
        self.assertFalse(NR(0, 1, 0, (True, True)).issubset(NR(0, 1, 0, B)))
        self.assertTrue(NR(0, 1, 0, (True, False)).issubset(NR(0, 1, 0, B)))
        self.assertFalse(NR(0, 1, 0, (False, True)).issubset(NR(0, 1, 0, B)))
        self.assertTrue(NR(0, 1, 0, (False, False)).issubset(NR(0, 1, 0, B)))

        B = False, True
        self.assertFalse(NR(0, 1, 0, (True, True)).issubset(NR(0, 1, 0, B)))
        self.assertFalse(NR(0, 1, 0, (True, False)).issubset(NR(0, 1, 0, B)))
        self.assertTrue(NR(0, 1, 0, (False, True)).issubset(NR(0, 1, 0, B)))
        self.assertTrue(NR(0, 1, 0, (False, False)).issubset(NR(0, 1, 0, B)))

        B = False, False
        self.assertFalse(NR(0, 1, 0, (True, True)).issubset(NR(0, 1, 0, B)))
        self.assertFalse(NR(0, 1, 0, (True, False)).issubset(NR(0, 1, 0, B)))
        self.assertFalse(NR(0, 1, 0, (False, True)).issubset(NR(0, 1, 0, B)))
        self.assertTrue(NR(0, 1, 0, (False, False)).issubset(NR(0, 1, 0, B)))

        # Continuous - discrete
        self.assertTrue(NR(0, None, 1).issubset(NR(None, None, 0)))
        self.assertTrue(NR(0, None, 1).issubset(NR(0, None, 0)))
        self.assertFalse(NR(0, None, 1).issubset(NR(None, 0, 0)))

        self.assertTrue(NR(0, None, -1).issubset(NR(None, None, 0)))
        self.assertFalse(NR(0, None, -1).issubset(NR(0, None, 0)))
        self.assertTrue(NR(0, None, -1).issubset(NR(None, 0, 0)))

        self.assertTrue(NR(0, 10, 1).issubset(NR(None, None, 0)))
        self.assertTrue(NR(0, 10, 1).issubset(NR(0, None, 0)))
        self.assertTrue(NR(0, 10, 1).issubset(NR(0, 10, 0)))

        self.assertFalse(NR(0, None, 0).issubset(NR(0, None, 1)))
        self.assertFalse(NR(None, 0, 0).issubset(NR(0, None, -1)))
        self.assertFalse(NR(0, 10, 0).issubset(NR(0, 10, 1)))

        # Discrete - discrete
        self.assertTrue(NR(0, 10, 2).issubset(NR(0, 10, 2)))
        self.assertTrue(NR(0, 10, 2).issubset(NR(-2, 10, 2)))
        self.assertTrue(NR(0, 10, 2).issubset(NR(0, 12, 2)))
        self.assertFalse(NR(0, 10, 3).issubset(NR(0, 10, 2)))
        self.assertTrue(NR(0, 11, 2).issubset(NR(0, 10, 2)))
        self.assertFalse(NR(1, 10, 2).issubset(NR(0, 10, 2)))
        self.assertFalse(NR(0, 10, 2).issubset(NR(0, 10, 4)))
        self.assertTrue(NR(0, 10, 2).issubset(NR(0, 10, 1)))

        self.assertTrue(NR(10, 0, -2).issubset(NR(10, 0, -2)))
        self.assertTrue(NR(10, 0, -2).issubset(NR(10, -2, -2)))
        self.assertTrue(NR(10, 0, -2).issubset(NR(12, 0, -2)))
        self.assertFalse(NR(10, 0, -3).issubset(NR(10, 0, -2)))
        self.assertTrue(NR(10, 1, -2).issubset(NR(10, 0, -2)))
        self.assertTrue(NR(8, 0, -2).issubset(NR(10, 0, -2)))
        self.assertFalse(NR(10, 0, -2).issubset(NR(10, 0, -4)))
        self.assertTrue(NR(10, 0, -2).issubset(NR(10, 0, -1)))

        # Scalar-discrete
        self.assertTrue(NR(5, 5, 0).issubset(NR(0, 10, 1)))
        self.assertFalse(NR(15, 15, 0).issubset(NR(0, 10, 1)))

    def test_lcm(self):
        self.assertEqual(NR(None, None, 0)._step_lcm((NR(0, 1, 0),)), 0)
        self.assertEqual(NR(None, None, 0)._step_lcm((NR(0, 0, 0),)), 1)
        self.assertEqual(NR(0, None, 3)._step_lcm((NR(0, None, 1),)), 3)
        self.assertEqual(NR(0, None, 3)._step_lcm((NR(0, None, 0),)), 3)
        self.assertEqual(NR(0, None, 0)._step_lcm((NR(0, None, 1),)), 1)
        self.assertEqual(NR(0, None, 3)._step_lcm((NR(0, None, 2),)), 6)
        self.assertEqual(NR(0, None, 18)._step_lcm((NR(0, None, 12),)), 36)
        self.assertEqual(NR(0, None, 3)._step_lcm((NR(0, None, 2), NR(0, None, 5))), 30)
        self.assertEqual(
            NR(0, None, 3)._step_lcm((NR(0, None, 2), NR(0, None, 10))), 30
        )

    def test_range_difference(self):
        self.assertEqual(
            NR(0, None, 1).range_difference([NR(1, None, 0)]), [NR(0, 0, 0)]
        )
        self.assertEqual(
            NR(0, None, 1).range_difference([NR(0, 0, 0)]), [NR(1, None, 1)]
        )
        self.assertEqual(
            NR(0, None, 2).range_difference([NR(10, None, 3)]),
            [NR(0, None, 6), NR(2, None, 6), NR(4, 4, 0)],
        )
        with self.assertRaisesRegex(ValueError, "Unknown range type, list"):
            NR(0, None, 0).range_difference([[0]])

        # test relatively prime ranges that don't expand to all offsets
        self.assertEqual(
            NR(0, 7, 2).range_difference([NR(6, None, 10)]),
            [NR(0, 0, 0), NR(2, 2, 0), NR(4, 4, 0)],
        )

        # test ranges running in the other direction
        self.assertEqual(
            NR(10, 0, -1).range_difference([NR(7, 4, -2)]),
            [NR(10, 0, -2), NR(1, 3, 2), NR(9, 9, 0)],
        )
        self.assertEqual(
            NR(0, None, -1).range_difference([NR(-10, 10, 0)]), [NR(-11, None, -1)]
        )

        # Test non-overlapping ranges
        self.assertEqual(NR(0, 4, 0).range_difference([NR(5, 10, 0)]), [NR(0, 4, 0)])
        self.assertEqual(NR(5, 10, 0).range_difference([NR(0, 4, 0)]), [NR(5, 10, 0)])

        # Test continuous ranges

        # Subtracting a closed range from a closed range should
        # result in an open range.
        self.assertEqual(
            NR(0, None, 0).range_difference([NR(5, None, 0)]), [NR(0, 5, 0, '[)')]
        )
        self.assertEqual(
            NR(0, None, 0).range_difference([NR(5, 10, 0)]),
            [NR(0, 5, 0, '[)'), NR(10, None, 0, '(]')],
        )
        self.assertEqual(
            NR(None, 0, 0).range_difference([NR(-5, None, 0)]), [NR(None, -5, 0, '[)')]
        )
        self.assertEqual(
            NR(None, 0, 0).range_difference([NR(-5, 0, 0, '[]')]),
            [NR(None, -5, 0, '[)')],
        )
        self.assertEqual(
            NR(None, 0, 0).range_difference([NR(-5, 0, 0, '[)')]),
            [NR(None, -5, 0, '[)'), NR(0, 0, 0)],
        )
        self.assertEqual(
            NR(0, 10, 0).range_difference([NR(None, 5, 0, '[)')]), [NR(5, 10, 0, '[]')]
        )
        # Subtracting an open range from a closed range gives a closed
        # range
        self.assertEqual(
            NR(0, None, 0).range_difference([NR(5, 10, 0, '()')]),
            [NR(0, 5, 0, '[]'), NR(10, None, 0, '[]')],
        )
        # Subtracting a discrete range from a continuous range gives a
        # set of open continuous ranges
        self.assertEqual(
            NR(None, None, 0).range_difference([NR(5, 10, 5)]),
            [NR(None, 5, 0, '[)'), NR(5, 10, 0, '()'), NR(10, None, 0, '(]')],
        )
        self.assertEqual(
            NR(-10, 20, 0).range_difference([NR(5, 10, 5)]),
            [NR(-10, 5, 0, '[)'), NR(5, 10, 0, '()'), NR(10, 20, 0, '(]')],
        )
        self.assertEqual(
            NR(-10, 20, 0, "()").range_difference([NR(5, 10, 5)]),
            [NR(-10, 5, 0, '()'), NR(5, 10, 0, '()'), NR(10, 20, 0, '()')],
        )
        self.assertEqual(
            NR(-3, 3, 0).range_difference([NR(0, None, 5), NR(0, None, -5)]),
            [NR(-3, 0, 0, '[)'), NR(0, 3, 0, '(]')],
        )

        # Disjoint ranges...
        a = NR(0.25, 10, 1)
        self.assertEqual(a.range_difference([NR(0.5, 20, 1)]), [a])
        self.assertEqual(
            a.range_difference([NR(0.5, 20, 2)]), [NR(0.25, 8.25, 2), NR(1.25, 9.25, 2)]
        )
        a = NR(0, 100, 2)
        self.assertEqual(
            a.range_difference([NR(1, 100, 4)]), [NR(0, 100, 4), NR(2, 98, 4)]
        )
        a = NR(0, None, 2)
        self.assertEqual(
            a.range_difference([NR(1, None, 4)]), [NR(0, None, 4), NR(2, None, 4)]
        )
        a = NR(0.25, None, 1)
        self.assertEqual(a.range_difference([NR(0.5, None, 1)]), [a])

        # open/closed infinite ranges
        a = NR(None, None, 0)
        self.assertEqual(
            a.range_difference([NR(None, None, 0, "()")]),
            [NR(-_inf, -_inf, 0), NR(_inf, _inf, 0)],
        )
        self.assertEqual(
            a.range_difference([NR(None, None, 0, "()"), NR(None, None, 0, "[)")]),
            [NR(_inf, _inf, 0)],
        )

        # And the one thing we don't support:
        with self.assertRaisesRegex(
            RangeDifferenceError,
            'We do not support subtracting an '
            r'infinite discrete range \[0:inf\] from an infinite '
            r'continuous range \[-inf..inf\]',
        ):
            NR(None, None, 0).range_difference([NR(0, None, 1)])

    def test_range_intersection(self):
        self.assertEqual(
            NR(0, None, 1).range_intersection([NR(1, None, 0)]), [NR(1, None, 1)]
        )
        self.assertEqual(
            NR(0, None, 1).range_intersection([NR(0, 0, 0)]), [NR(0, 0, 0)]
        )
        self.assertEqual(
            NR(0, None, 1).range_intersection([NR(0.5, 1.5, 0)]), [NR(1, 1, 0)]
        )
        self.assertEqual(
            NR(0, None, 2).range_intersection([NR(1, None, 3)]), [NR(4, None, 6)]
        )
        with self.assertRaisesRegex(ValueError, "Unknown range type, list"):
            NR(0, None, 0).range_intersection([[0]])

        # Test non-overlapping ranges
        self.assertEqual(NR(0, 4, 0).range_intersection([NR(5, 10, 0)]), [])
        self.assertEqual(NR(5, 10, 0).range_intersection([NR(0, 4, 0)]), [])
        self.assertEqual(NR(0, 4, 0).range_intersection([NNR('a')]), [])

        # test ranges running in the other direction
        self.assertEqual(
            NR(10, 0, -1).range_intersection([NR(7, 4, -2)]), [NR(5, 7, 2)]
        )
        self.assertEqual(
            NR(10, 0, -1).range_intersection([NR(7, None, -2)]), [NR(1, 7, 2)]
        )
        self.assertEqual(
            NR(0, None, -1).range_intersection([NR(None, -10, 0)]), [NR(-10, None, -1)]
        )

        # Test continuous ranges
        self.assertEqual(NR(0, 5, 0).range_intersection([NR(5, 10, 0)]), [NR(5, 5, 0)])
        self.assertEqual(
            NR(0, None, 0).range_intersection([NR(5, None, 0)]), [NR(5, None, 0)]
        )

        # Disjoint ranges...
        a = NR(0.25, 10, 1)
        self.assertEqual(a.range_intersection([NR(0.5, 20, 1)]), [])
        self.assertEqual(a.range_intersection([NR(0.5, 20, 2)]), [])
        a = NR(0, 100, 2)
        self.assertEqual(a.range_intersection([NR(1, 100, 4)]), [])
        a = NR(0, None, 2)
        self.assertEqual(a.range_intersection([NR(1, None, 4)]), [])
        a = NR(0.25, None, 1)
        self.assertEqual(a.range_intersection([NR(0.5, None, 1)]), [])

    def test_pickle(self):
        a = NR(0, 100, 5)
        b = pickle.loads(pickle.dumps(a))
        self.assertIsNot(a, b)
        self.assertEqual(a, b)


class TestAnyRange(unittest.TestCase):
    def test_str(self):
        self.assertEqual(str(AnyRange()), '[*]')

    def test_range_relational(self):
        a = AnyRange()
        b = AnyRange()
        self.assertTrue(a.issubset(b))
        self.assertEqual(a, a)
        self.assertEqual(a, b)

        c = NR(None, None, 0)
        self.assertFalse(a.issubset(c))
        self.assertTrue(c.issubset(b))
        self.assertNotEqual(a, c)
        self.assertNotEqual(c, a)

    def test_contains(self):
        a = AnyRange()
        self.assertIn(None, a)
        self.assertIn(0, a)
        self.assertIn('a', a)

    def test_range_difference(self):
        self.assertEqual(AnyRange().range_difference([NR(0, None, 1)]), [AnyRange()])
        self.assertEqual(NR(0, None, 1).range_difference([AnyRange()]), [])
        self.assertEqual(AnyRange().range_difference([AnyRange()]), [])

    def test_range_intersection(self):
        self.assertEqual(
            AnyRange().range_intersection([NR(0, None, 1)]), [NR(0, None, 1)]
        )
        self.assertEqual(
            NR(0, None, 1).range_intersection([AnyRange()]), [NR(0, None, 1)]
        )
        self.assertEqual(
            NR(0, None, -1).range_intersection([AnyRange()]), [NR(0, None, -1)]
        )

    def test_info_methods(self):
        a = AnyRange()
        self.assertFalse(a.isdiscrete())
        self.assertFalse(a.isfinite())

    def test_pickle(self):
        a = AnyRange()
        b = pickle.loads(pickle.dumps(a))
        self.assertIsNot(a, b)
        self.assertEqual(a, b)


class TestNonNumericRange(unittest.TestCase):
    def test_str(self):
        a = NNR('a')
        aa = NNR('a')
        b = NNR(None)
        self.assertEqual(str(a), '{a}')
        self.assertEqual(str(aa), '{a}')
        self.assertEqual(str(b), '{None}')

    def test_range_relational(self):
        a = NNR('a')
        aa = NNR('a')
        b = NNR(None)
        self.assertTrue(a.issubset(aa))
        self.assertFalse(a.issubset(b))
        self.assertEqual(a, a)
        self.assertEqual(a, aa)
        self.assertNotEqual(a, b)

        c = NR(None, None, 0)
        self.assertFalse(a.issubset(c))
        self.assertFalse(c.issubset(b))
        self.assertNotEqual(a, c)
        self.assertNotEqual(c, a)

    def test_contains(self):
        a = NNR('a')
        b = NNR(None)
        self.assertIn('a', a)
        self.assertNotIn(0, a)
        self.assertNotIn(None, a)
        self.assertNotIn('a', b)
        self.assertNotIn(0, b)
        self.assertIn(None, b)

    def test_range_difference(self):
        a = NNR('a')
        b = NNR(None)
        self.assertEqual(a.range_difference([NNR('a')]), [])
        self.assertEqual(a.range_difference([b]), [NNR('a')])

    def test_range_intersection(self):
        a = NNR('a')
        b = NNR(None)
        self.assertEqual(a.range_intersection([b]), [])
        self.assertEqual(a.range_intersection([NNR('a')]), [NNR('a')])

    def test_info_methods(self):
        a = NNR('a')
        self.assertTrue(a.isdiscrete())
        self.assertTrue(a.isfinite())

    def test_pickle(self):
        a = NNR('a')
        b = pickle.loads(pickle.dumps(a))
        self.assertIsNot(a, b)
        self.assertEqual(a, b)


class TestRangeProduct(unittest.TestCase):
    def test_str(self):
        a = RP([[NR(0, 10, 1)], [NR(0, 10, 0), NNR('a')]])
        self.assertEqual(str(a), '<[0:10], ([0..10], {a})>')

    def test_range_relational(self):
        a = RP([[NR(0, 10, 1)], [NR(0, 10, 0), NNR('a')]])
        aa = RP([[NR(0, 10, 1)], [NR(0, 10, 0), NNR('a')]])
        b = RP([[NR(0, 10, 1)], [NR(0, 10, 0), NNR('a'), NNR('b')]])
        c = RP([[NR(0, 10, 1)], [NR(0, 10, 0), NNR('b')]])
        d = RP([[NR(0, 10, 0)], [NR(0, 10, 0), NNR('a')]])
        d = RP([[NR(0, 10, 0)], [AnyRange()]])

        self.assertTrue(a.issubset(aa))
        self.assertTrue(a.issubset(b))
        self.assertFalse(a.issubset(c))
        self.assertTrue(a.issubset(d))

        self.assertFalse(a.issubset(NNR('a')))
        self.assertFalse(a.issubset(NR(None, None, 0)))
        self.assertTrue(a.issubset(AnyRange()))

    def test_contains(self):
        a = NNR('a')
        b = NR(0, 5, 0)
        c = NR(5, 10, 1)
        x = RP([[a], [b, c]])
        self.assertNotIn('a', x)
        self.assertNotIn(0, x)
        self.assertNotIn(None, x)
        self.assertIn(('a', 0), x)
        self.assertIn(('a', 6), x)
        self.assertNotIn(('a', 6.5), x)

    def test_equality(self):
        a = NNR('a')
        b = NR(0, 5, 0)
        c = NR(5, 10, 1)
        x = RP([[a], [b, c]])
        y = RP([[a], [c]])
        self.assertEqual(x, x)
        self.assertNotEqual(x, y)

    def test_isdisjoint(self):
        a = NNR('a')
        b = NR(0, 5, 0)
        c = NR(5, 10, 1)
        x = RP([[a], [b, c]])
        y = RP([[a], [c]])
        z = RP([[a], [b], [c]])
        w = RP([[AnyRange()], [b]])
        self.assertFalse(x.isdisjoint(x))
        self.assertFalse(x.isdisjoint(y))
        self.assertTrue(x.isdisjoint(z))
        self.assertFalse(x.isdisjoint(w))
        self.assertTrue(x.isdisjoint(a))
        self.assertFalse(y.isdisjoint(w))
        self.assertFalse(x.isdisjoint(AnyRange()))
        v = RP([[AnyRange()], [NR(0, 5, 0, (False, False))]])
        self.assertTrue(y.isdisjoint(v))

    def test_range_difference(self):
        a = NNR('a')
        b = NR(0, 5, 0)
        b1 = NR(0, 5, 0, '[)')  # Note: b & c overlap, so [b,c]-c != b
        c = NR(5, 10, 1)
        x = RP([[a], [b, c]])
        y = RP([[a], [c]])
        z = RP([[a], [b], [c]])
        w = RP([list(Any.ranges()), [b]])
        self.assertEqual(x.range_difference([x]), [])
        self.assertEqual(x.range_difference([y]), [RP([[a], [b1]])])
        self.assertEqual(x.range_difference([z]), [x])
        self.assertEqual(x.range_difference(Any.ranges()), [])
        self.assertEqual(x.range_difference([w]), [RP([[a], [NR(6, 10, 1)]])])
        v = RP([[AnyRange()], [NR(0, 5, 0, (False, False))]])
        self.assertEqual(y.range_difference([v]), [y])

    def test_range_intersection(self):
        a = NNR('a')
        b = NR(0, 5, 0)
        c = NR(5, 10, 1)
        x = RP([[a], [b, c]])
        y = RP([[a], [c]])
        z = RP([[a], [b], [c]])
        w = RP([list(Any.ranges()), [b]])
        self.assertEqual(x.range_intersection([x]), [x])
        self.assertEqual(x.range_intersection([y]), [y])
        self.assertEqual(x.range_intersection([z]), [])
        self.assertEqual(x.range_intersection(Any.ranges()), [x])
        self.assertEqual(x.range_intersection([w]), [RP([[a], [b]])])
        self.assertEqual(y.range_intersection([w]), [RP([[a], [NR(5, 5, 0)]])])
        v = RP([[AnyRange()], [NR(0, 5, 0, (False, False))]])
        self.assertEqual(y.range_intersection([v]), [])

    def test_info_methods(self):
        a = NNR('a')
        b = NR(0, 5, 0)
        c = NR(5, 10, 1)
        x = RP([[a], [b, c]])
        y = RP([[a], [c]])
        self.assertFalse(x.isdiscrete())
        self.assertFalse(x.isfinite())
        self.assertTrue(y.isdiscrete())
        self.assertTrue(y.isfinite())

    def test_pickle(self):
        a = NNR('a')
        b = NR(0, 5, 0)
        c = NR(5, 10, 1)
        x = RP([[a], [b, c]])
        y = RP([[a], [c]])

        xx = pickle.loads(pickle.dumps(x))
        self.assertIsNot(x, xx)
        self.assertEqual(x, xx)

        yy = pickle.loads(pickle.dumps(y))
        self.assertIsNot(y, yy)
        self.assertEqual(y, yy)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Transformation to strip variable bounds from a model."""

from pyomo.common.collections import ComponentMap
from pyomo.common.config import (
    ConfigBlock,
    ConfigValue,
    document_kwargs_from_configdict,
)
from pyomo.core.base.transformation import TransformationFactory
from pyomo.core.base.var import Var
from pyomo.core.base.set_types import Reals
from pyomo.core.plugins.transform.hierarchy import NonIsomorphicTransformation


@TransformationFactory.register(
    'contrib.strip_var_bounds', doc="Strip bounds from variables."
)
@document_kwargs_from_configdict('CONFIG')
class VariableBoundStripper(NonIsomorphicTransformation):
    """Strip bounds from variables.

    Keyword arguments below are specified for the ``apply_to`` and
    ``create_using`` functions.

    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "strip_domains",
        ConfigValue(
            default=True,
            domain=bool,
            description="strip the domain for discrete variables as well",
        ),
    )
    CONFIG.declare(
        "reversible",
        ConfigValue(
            default=False,
            domain=bool,
            description="Whether the bound stripping will be temporary. "
            "If so, store information for reversion.",
        ),
    )

    def _apply_to(self, instance, **kwds):
        config = self.CONFIG(kwds)
        if config.reversible:
            if any(
                hasattr(instance, map_name)
                for map_name in [
                    '_tmp_var_bound_strip_lb',
                    '_tmp_var_bound_strip_ub',
                    '_tmp_var_bound_strip_domain',
                ]
            ):
                raise RuntimeError(
                    'Variable stripping reversion component maps already '
                    'exist. Did you already apply a temporary transformation '
                    'without a subsequent reversion?'
                )
            # Component maps to store data for reversion.
            instance._tmp_var_bound_strip_lb = ComponentMap()
            instance._tmp_var_bound_strip_ub = ComponentMap()
            instance._tmp_var_bound_strip_domain = ComponentMap()
        for var in instance.component_data_objects(ctype=Var):
            if config.strip_domains and not var.domain == Reals:
                if config.reversible:
                    instance._tmp_var_bound_strip_domain[var] = var.domain
                var.domain = Reals
            if var.has_lb():
                if config.reversible:
                    instance._tmp_var_bound_strip_lb[var] = var.lb
                var.setlb(None)
            if var.has_ub():
                if config.reversible:
                    instance._tmp_var_bound_strip_ub[var] = var.ub
                var.setub(None)

    def revert(self, instance):
        """Revert variable bounds and domains changed by the transformation."""
        for var in instance.component_data_objects(ctype=Var, descend_into=True):
            if var in instance._tmp_var_bound_strip_lb:
                var.setlb(instance._tmp_var_bound_strip_lb[var])
            if var in instance._tmp_var_bound_strip_ub:
                var.setub(instance._tmp_var_bound_strip_ub[var])
            if var in instance._tmp_var_bound_strip_domain:
                var.domain = instance._tmp_var_bound_strip_domain[var]
        del instance._tmp_var_bound_strip_lb
        del instance._tmp_var_bound_strip_ub
        del instance._tmp_var_bound_strip_domain



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import numpy as np
from pyomo.common.dependencies import mpi4py
from pyomo.contrib.pynumero.sparse.mpi_block_vector import MPIBlockVector


def main():
    comm = mpi4py.MPI.COMM_WORLD
    rank = comm.Get_rank()

    owners = [2, 0, 1, -1]
    x = MPIBlockVector(4, rank_owner=owners, mpi_comm=comm)
    x.set_block(owners.index(rank), np.ones(3) * (rank + 1))
    x.set_block(3, np.array([1, 2, 3]))

    y = MPIBlockVector(4, rank_owner=owners, mpi_comm=comm)
    y.set_block(owners.index(rank), np.ones(3) * (rank + 1))
    y.set_block(3, np.array([1, 2, 3]))

    z1: MPIBlockVector = x + y  # add x and y
    z2 = x.dot(y)  # dot product
    z3 = np.abs(x).max()  # infinity norm

    z1_local = z1.make_local_copy()
    if rank == 0:
        print(z1_local.flatten())
        print(z2)
        print(z3)

    return z1_local, z2, z3


if __name__ == '__main__':
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Test separation problem construction methods.
"""

import logging
import pyomo.common.unittest as unittest
from pyomo.common.log import LoggingIntercept

from pyomo.common.collections import Bunch
from pyomo.common.dependencies import numpy as np, numpy_available, scipy_available
from pyomo.core.base import ConcreteModel, Constraint, Objective, Param, Var
from pyomo.core.expr import exp, RangedExpression, value
from pyomo.core.expr.compare import assertExpressionsEqual

from pyomo.contrib.pyros.master_problem_methods import (
    MasterProblemData,
    add_scenario_block_to_master_problem,
)
from pyomo.contrib.pyros.separation_problem_methods import (
    construct_separation_problem,
    group_ss_ineq_constraints_by_priority,
    SeparationProblemData,
    initialize_separation,
)
from pyomo.contrib.pyros.uncertainty_sets import (
    BoxSet,
    FactorModelSet,
    DiscreteScenarioSet,
)
from pyomo.contrib.pyros.util import (
    ModelData,
    preprocess_model_data,
    ObjectiveType,
    VariablePartitioning,
)

if not (numpy_available and scipy_available):
    raise unittest.SkipTest("Packages numpy and scipy must both be available.")


logger = logging.getLogger(__name__)


def build_simple_model_data(objective_focus="worst_case", uncertainty_set=None):
    """
    Build simple model data object for master problem construction.
    """
    m = ConcreteModel()
    m.u = Param(initialize=0.5, mutable=True)
    m.u2 = Param(initialize=0, mutable=True)
    m.x1 = Var(bounds=[-1000, 1000])
    m.x2 = Var(bounds=[-1000, 1000])
    m.x3 = Var(bounds=[-1000, 1000])
    m.con = Constraint(expr=exp(m.u - 1) - m.x1 - m.x2 * m.u - m.x3 * m.u**2 <= 0)

    # this makes x2 nonadjustable
    m.eq_con = Constraint(expr=m.x2 - 1 == 0)

    m.obj = Objective(expr=m.x1 + m.x2 / 2 + m.x3 / 3 + m.u + m.u2)

    if uncertainty_set is None:
        uncertainty_set = BoxSet([[0, 1], [0, 0]])

    config = Bunch(
        uncertain_params=[m.u, m.u2],
        objective_focus=ObjectiveType[objective_focus],
        decision_rule_order=1,
        progress_logger=logger,
        nominal_uncertain_param_vals=[0.5, 0],
        uncertainty_set=uncertainty_set,
        separation_priority_order=dict(con=2),
    )
    model_data = ModelData(original_model=m, timing=None, config=config)
    user_var_partitioning = VariablePartitioning(
        first_stage_variables=[m.x1],
        second_stage_variables=[m.x2, m.x3],
        state_variables=[],
    )

    preprocess_model_data(model_data, user_var_partitioning)

    return model_data


class TestConstructSeparationProblem(unittest.TestCase):
    """
    Test method for construction of separation problem.
    """

    def test_construct_separation_problem_nonadj_components(self):
        """
        Check first-stage variables and constraints of the
        separation problem are fixed and deactivated,
        respectively.
        """
        model_data = build_simple_model_data(objective_focus="worst_case")
        separation_model = construct_separation_problem(model_data)

        # check nonadjustable components fixed/deactivated
        self.assertTrue(separation_model.user_model.x1.fixed)
        self.assertTrue(separation_model.first_stage.epigraph_var.fixed)
        for indexed_var in separation_model.first_stage.decision_rule_vars:
            for dr_var in indexed_var.values():
                self.assertTrue(dr_var.fixed, msg=f"DR var {dr_var.name!r} not fixed")

        # first-stage equality constraints should be inactive
        self.assertFalse(separation_model.user_model.eq_con.active)
        for coeff_con in separation_model.first_stage.coefficient_matching_cons:
            self.assertFalse(
                coeff_con.active,
                msg=f"Coefficient matching constraint {coeff_con.name!r} active.",
            )

    def test_construct_separation_problem_ss_ineq_cons(self):
        """
        Check second-stage inequality constraints are deactivated
        and replaced with objectives, as appropriate.
        """
        model_data = build_simple_model_data(objective_focus="worst_case")
        separation_model = construct_separation_problem(model_data)

        # check expression of second-stage ineq cons correct
        # check these individually
        # (i.e. uncertain params have been replaced)
        m = separation_model.user_model
        u1_var = separation_model.uncertainty.uncertain_param_var_list[0]
        u2_var = separation_model.uncertainty.uncertain_param_var_list[1]
        assertExpressionsEqual(
            self,
            separation_model.second_stage.inequality_cons["epigraph_con"].expr,
            (
                m.x1
                + m.x2 / 2
                + m.x3 / 3
                + u1_var
                + u2_var
                - separation_model.first_stage.epigraph_var
                <= 0
            ),
        )

        self.assertFalse(
            separation_model.second_stage.inequality_cons["epigraph_con"].active
        )
        self.assertFalse(
            m.con.active,
            separation_model.second_stage.inequality_cons[
                "ineq_con_con_upper_bound_con"
            ].active,
        )
        self.assertFalse(
            m.con.active,
            separation_model.second_stage.inequality_cons[
                "var_x3_certain_lower_bound_con"
            ].active,
        )
        self.assertFalse(
            m.con.active,
            separation_model.second_stage.inequality_cons[
                "var_x3_certain_upper_bound_con"
            ].active,
        )

        # check second-stage ineq con expressions match obj expressions
        # (loop through the con to obj map)
        self.assertEqual(
            len(separation_model.second_stage_ineq_con_to_obj_map),
            len(separation_model.second_stage.inequality_cons),
        )
        for ineq_con, obj in separation_model.second_stage_ineq_con_to_obj_map.items():
            assertExpressionsEqual(self, ineq_con.body - ineq_con.upper, obj.expr)

    def test_construct_separation_problem_ss_eq_and_dr_cons(self):
        """
        Check second-stage and DR equations are appropriately handled
        by the separation problems.
        """
        # check DR equation is active
        model_data = build_simple_model_data(objective_focus="worst_case")
        separation_model = construct_separation_problem(model_data)

        self.assertTrue(separation_model.second_stage.decision_rule_eqns[0].active)

        u1_var = separation_model.uncertainty.uncertain_param_var_list[0]
        assertExpressionsEqual(
            self,
            separation_model.second_stage.decision_rule_eqns[0].expr,
            (
                separation_model.first_stage.decision_rule_vars[0][0]
                # only one effective uncertain parameter,
                # so only one affine DR term
                + u1_var * separation_model.first_stage.decision_rule_vars[0][1]
                - separation_model.user_model.x3
                == 0
            ),
        )

    def test_construct_separation_problem_uncertainty_components(self):
        """
        Test separation problem handles uncertain parameter variable
        components as expected.
        """
        model_data = build_simple_model_data(objective_focus="worst_case")
        separation_model = construct_separation_problem(model_data)
        uncertainty_blk = separation_model.uncertainty
        boxcon1, boxcon2 = uncertainty_blk.uncertainty_cons_list
        paramvar1, paramvar2 = uncertainty_blk.uncertain_param_var_list

        self.assertEqual(uncertainty_blk.auxiliary_var_list, [])
        self.assertEqual(len(uncertainty_blk.uncertainty_cons_list), 2)
        assertExpressionsEqual(
            self,
            boxcon1.expr,
            RangedExpression((np.int_(0), paramvar1, np.int_(1)), False),
        )
        assertExpressionsEqual(
            self,
            boxcon2.expr,
            RangedExpression((np.int_(0), paramvar2, np.int_(0)), False),
        )
        self.assertTrue(boxcon1.active)
        # constraints in fixed uncertain params should be deactivated
        self.assertFalse(boxcon2.active)

        # u, bounds [0, 1]
        self.assertFalse(paramvar1.fixed)
        # bounds [0, 0]; separation constructor should fix the Var
        self.assertTrue(paramvar2.fixed)

        self.assertEqual(paramvar1.bounds, (0, 1))
        self.assertEqual(paramvar2.bounds, (0, 0))

    def test_construct_separation_problem_uncertain_factor_param_components(self):
        """
        Test separation problem uncertainty components for uncertainty
        set requiring auxiliary variables.
        """
        model_data = build_simple_model_data(
            objective_focus="worst_case",
            uncertainty_set=FactorModelSet(
                origin=[1, 0], beta=1, number_of_factors=2, psi_mat=[[1, 2.5], [0, 1]]
            ),
        )
        separation_model = construct_separation_problem(model_data)
        uncertainty_blk = separation_model.uncertainty
        *matrix_product_cons, aux_sum_con = uncertainty_blk.uncertainty_cons_list
        paramvar1, paramvar2 = uncertainty_blk.uncertain_param_var_list
        auxvar1, auxvar2 = uncertainty_blk.auxiliary_var_list

        self.assertEqual(len(matrix_product_cons), 2)
        self.assertTrue(matrix_product_cons[0].active)
        self.assertTrue(matrix_product_cons[1].active)
        self.assertTrue(aux_sum_con.active)
        assertExpressionsEqual(
            self, aux_sum_con.expr, RangedExpression((-2, auxvar1 + auxvar2, 2), False)
        )
        assertExpressionsEqual(
            self, matrix_product_cons[0].expr, auxvar1 + 2.5 * auxvar2 + 1 == paramvar1
        )
        assertExpressionsEqual(
            self, matrix_product_cons[1].expr, 0.0 * auxvar1 + auxvar2 == paramvar2
        )

        # none of the vars should be fixed
        self.assertFalse(paramvar1.fixed)
        self.assertFalse(paramvar2.fixed)
        self.assertFalse(auxvar1.fixed)
        self.assertFalse(auxvar2.fixed)

        # factor set auxiliary variables
        self.assertEqual(auxvar1.bounds, (-1, 1))
        self.assertEqual(auxvar2.bounds, (-1, 1))

        # factor set bounds are tighter
        self.assertEqual(paramvar1.bounds, (-2.5, 4.5))
        self.assertEqual(paramvar2.bounds, (-1.0, 1.0))

        # check decision rule equation
        assertExpressionsEqual(
            self,
            separation_model.second_stage.decision_rule_eqns[0].expr,
            (
                separation_model.first_stage.decision_rule_vars[0][0]
                + paramvar1 * separation_model.first_stage.decision_rule_vars[0][1]
                + paramvar2 * separation_model.first_stage.decision_rule_vars[0][2]
                - separation_model.user_model.x3
                == 0
            ),
        )


class TestGroupSecondStageIneqConsByPriority(unittest.TestCase):
    def test_group_ss_ineq_constraints_by_priority(self):
        model_data = build_simple_model_data()
        model_data.config.progress_logger = logging.getLogger(__name__)
        separation_model = construct_separation_problem(model_data)

        # build mock separation data-like object
        # since we are testing only the grouping method
        separation_data = Bunch(
            separation_model=separation_model,
            separation_priority_order=model_data.separation_priority_order,
            config=model_data.config,
        )

        priority_groups = group_ss_ineq_constraints_by_priority(separation_data)

        self.assertEqual(list(priority_groups.keys()), [2, 0])
        ss_ineq_cons = separation_model.second_stage.inequality_cons
        self.assertEqual(
            priority_groups[2], [ss_ineq_cons["ineq_con_con_upper_bound_con"]]
        )
        self.assertEqual(
            priority_groups[0],
            [
                ss_ineq_cons["var_x3_certain_lower_bound_con"],
                ss_ineq_cons["var_x3_certain_upper_bound_con"],
                ss_ineq_cons["epigraph_con"],
            ],
        )


class TestInitializeSeparation(unittest.TestCase):
    """
    Tests for separation subproblem initialization.
    """

    def test_initialize_separation(self):
        model_data = build_simple_model_data(
            objective_focus="worst_case",
            uncertainty_set=FactorModelSet(
                # note: origin is chosen to be different
                #       from the nominal value so that
                #       initialization of auxiliary uncertain
                #       parameters is meaningfully tested
                origin=[1, 0.5],
                psi_mat=[[1, 1], [1, 0]],
                beta=0.5,
                number_of_factors=2,
            ),
        )

        master_data = MasterProblemData(model_data)
        nom_scenario_blk = master_data.master_model.scenarios[0, 0]
        nom_scenario_blk.user_model.x1.set_value(10)
        nom_scenario_blk.user_model.x2.set_value(1)
        nom_scenario_blk.user_model.x3.set_value(5)
        nom_scenario_blk.first_stage.decision_rule_vars[0][0].set_value(5)
        nom_scenario_blk.first_stage.epigraph_var.set_value(
            value(nom_scenario_blk.full_objective)
        )

        # set up new scenario block
        new_master_param_realization = [1.5, 1]
        add_scenario_block_to_master_problem(
            master_model=master_data.master_model,
            scenario_idx=(1, 0),
            param_realization=new_master_param_realization,
            from_block=nom_scenario_blk,
            clone_first_stage_components=False,
        )
        new_scenario_blk = master_data.master_model.scenarios[1, 0]
        new_scenario_blk.first_stage.epigraph_var.set_value(
            # objective for new block is higher,
            # so update the epigraph variable
            value(new_scenario_blk.full_objective)
        )
        # different value for the adjustable variable
        # so we also adjust the DR variables
        # to ensure the DR equations are satisfied
        new_scenario_blk.user_model.x3.set_value(6)
        new_scenario_blk.first_stage.decision_rule_vars[0][0].set_value(4.5)
        new_scenario_blk.first_stage.decision_rule_vars[0][1].set_value(1)

        separation_data = SeparationProblemData(model_data)

        ss_ineq_con_to_maximize = (
            separation_data.separation_model.second_stage.inequality_cons[
                "epigraph_con"
            ]
        )
        separation_data.points_added_to_master[1, 0] = new_master_param_realization
        separation_data.auxiliary_values_for_master_points[(1, 0)] = (
            model_data.config.uncertainty_set.compute_auxiliary_uncertain_param_vals(
                point=new_master_param_realization, solver=None
            )
        )

        with LoggingIntercept(module=__name__, level=logging.DEBUG) as LOG:
            initialize_separation(
                ss_ineq_con_to_maximize=ss_ineq_con_to_maximize,
                separation_data=separation_data,
                master_data=master_data,
            )

        # the constraint violation being maximized for
        # has a higher value in the newer master block
        init_from_scenario_blk = new_scenario_blk

        # all active constraints of the separation problem
        # should be satisfied post initialization, so expect
        # no logging messages stating otherwise
        log_output = LOG.getvalue()
        self.assertFalse(log_output, "DEBUG-level log output should be empty.")

        sep_usr_blk = separation_data.separation_model.user_model
        sep_model = separation_data.separation_model

        # first-stage variables added by PyROS
        self.assertEqual(
            value(sep_model.first_stage.epigraph_var),
            value(init_from_scenario_blk.first_stage.epigraph_var),
        )
        self.assertEqual(
            value(sep_model.first_stage.decision_rule_vars[0][0]),
            value(init_from_scenario_blk.first_stage.decision_rule_vars[0][0]),
        )
        self.assertEqual(
            value(sep_model.first_stage.decision_rule_vars[0][1]),
            value(init_from_scenario_blk.first_stage.decision_rule_vars[0][1]),
        )
        self.assertEqual(
            value(sep_model.first_stage.decision_rule_vars[0][2]),
            value(init_from_scenario_blk.first_stage.decision_rule_vars[0][2]),
        )

        # variables of original user model
        self.assertEqual(
            value(sep_usr_blk.x1), value(init_from_scenario_blk.user_model.x1)
        )
        self.assertEqual(
            value(sep_usr_blk.x2), value(init_from_scenario_blk.user_model.x2)
        )
        self.assertEqual(
            value(sep_usr_blk.x3), value(init_from_scenario_blk.user_model.x3)
        )

        # main uncertain parameter variables
        self.assertEqual(
            value(sep_model.uncertainty.uncertain_param_indexed_var[0]),
            value(init_from_scenario_blk.uncertain_params[0]),
        )
        self.assertEqual(
            value(sep_model.uncertainty.uncertain_param_indexed_var[1]),
            value(init_from_scenario_blk.uncertain_params[1]),
        )

        # auxiliary uncertain parameter variables
        expected_aux_var_vals = separation_data.auxiliary_values_for_master_points[
            (1, 0)
        ]
        self.assertEqual(
            value(sep_model.uncertainty.auxiliary_var_list[0]), expected_aux_var_vals[0]
        )
        self.assertEqual(
            value(sep_model.uncertainty.auxiliary_var_list[1]), expected_aux_var_vals[1]
        )

    def test_initialize_separation_infeasibility_logging(self):
        """
        Test initialization of a separation problem for which
        one of the active separation model constraints is
        not satisfied.
        """
        model_data = build_simple_model_data(
            objective_focus="worst_case",
            uncertainty_set=FactorModelSet(
                # note: origin is chosen to be different
                #       from the nominal value so that
                #       initialization of auxiliary uncertain
                #       parameters is meaningfully tested
                origin=[1, 0.5],
                psi_mat=[[1, 1], [1, 0]],
                beta=0.5,
                number_of_factors=2,
            ),
        )

        master_data = MasterProblemData(model_data)
        nom_scenario_blk = master_data.master_model.scenarios[0, 0]
        nom_scenario_blk.user_model.x1.set_value(10)
        nom_scenario_blk.user_model.x2.set_value(1)

        # this results in a violation of the DR equality constraint
        nom_scenario_blk.user_model.x3.set_value(5 + 1.5e-5)
        nom_scenario_blk.first_stage.decision_rule_vars[0][0].set_value(5)

        nom_scenario_blk.first_stage.epigraph_var.set_value(
            value(nom_scenario_blk.full_objective)
        )

        separation_data = SeparationProblemData(model_data)
        ss_ineq_con_to_maximize = (
            separation_data.separation_model.second_stage.inequality_cons[
                "epigraph_con"
            ]
        )
        with LoggingIntercept(module=__name__, level=logging.DEBUG) as LOG:
            initialize_separation(
                ss_ineq_con_to_maximize=ss_ineq_con_to_maximize,
                separation_data=separation_data,
                master_data=master_data,
            )
        log_output = LOG.getvalue()
        log_output_lines = log_output.split("\n")[:-1]
        self.assertEqual(
            len(log_output_lines),
            1,
            "Expected DEBUG-level output for separation problem initialization "
            "test to have exactly 1 line.",
        )
        self.assertRegex(
            log_output,
            r"Initial point for separation of .*violates the model constraint "
            r"'second_stage.decision_rule_eqns\[0\].*' by more than.*",
        )

    def test_initialize_separation_trivially_infeas_uncertainty(self):
        """
        Test error raised by separation problem initialization routine
        if the routine finds violated constraints depending
        only on fixed uncertain parameters.
        """
        model_data = build_simple_model_data(
            objective_focus="worst_case",
            # notice: the bounds mean the first uncertain
            # parameter will be fixed
            uncertainty_set=BoxSet(bounds=[[0.5, 0.5], [0, 1]]),
        )

        master_data = MasterProblemData(model_data)
        nom_scenario_blk = master_data.master_model.scenarios[0, 0]
        nom_scenario_blk.user_model.x1.set_value(10)
        nom_scenario_blk.user_model.x2.set_value(1)
        nom_scenario_blk.user_model.x3.set_value(5)
        nom_scenario_blk.first_stage.decision_rule_vars[0][0].set_value(5)
        nom_scenario_blk.first_stage.epigraph_var.set_value(
            value(nom_scenario_blk.full_objective)
        )

        separation_data = SeparationProblemData(model_data)
        # notice: the artificial point is such that the box constraint
        #         in the effectively certain parameter will be violated;
        #         this should trigger an exception
        separation_data.points_added_to_master[(0, 0)] = (1.5, 0)
        ss_ineq_con_to_maximize = (
            separation_data.separation_model.second_stage.inequality_cons[
                "epigraph_con"
            ]
        )
        exc_str = "Trivial infeasibility detected in the.*BoxSet.*constraints"
        with self.assertRaisesRegex(ValueError, exc_str):
            initialize_separation(
                ss_ineq_con_to_maximize=ss_ineq_con_to_maximize,
                separation_data=separation_data,
                master_data=master_data,
            )

    def test_initialize_separation_discrete_uncertainty(self):
        """
        Test initialization of a separation problem
        with a discrete uncertainty set.
        """
        model_data = build_simple_model_data(
            objective_focus="worst_case",
            uncertainty_set=DiscreteScenarioSet(scenarios=[[0.5, 0], [1, 0.5]]),
        )

        master_data = MasterProblemData(model_data)
        nom_scenario_blk = master_data.master_model.scenarios[0, 0]
        nom_scenario_blk.user_model.x1.set_value(10)
        nom_scenario_blk.user_model.x2.set_value(1)

        # this results in a violation of the DR equality constraint,
        # which is not logged, since the uncertainty set is discrete
        nom_scenario_blk.user_model.x3.set_value(5 + 1.1e-5)
        nom_scenario_blk.first_stage.decision_rule_vars[0][0].set_value(5)

        nom_scenario_blk.first_stage.epigraph_var.set_value(
            value(nom_scenario_blk.full_objective)
        )

        separation_data = SeparationProblemData(model_data)
        ss_ineq_con_to_maximize = (
            separation_data.separation_model.second_stage.inequality_cons[
                "epigraph_con"
            ]
        )

        sep_usr_blk = separation_data.separation_model.user_model
        sep_model = separation_data.separation_model

        # fix uncertain parameters to off-nominal value;
        # these should not be modified by the initialization
        off_nominal_scenario = model_data.config.uncertainty_set.scenarios[1]
        sep_model.uncertainty.uncertain_param_var_list[0].set_value(
            off_nominal_scenario[0]
        )
        sep_model.uncertainty.uncertain_param_var_list[1].set_value(
            off_nominal_scenario[1]
        )
        sep_model.uncertainty.uncertain_param_var_list[0].fix()
        sep_model.uncertainty.uncertain_param_var_list[1].fix()

        with LoggingIntercept(module=__name__, level=logging.DEBUG) as LOG:
            initialize_separation(
                ss_ineq_con_to_maximize=ss_ineq_con_to_maximize,
                separation_data=separation_data,
                master_data=master_data,
            )
        log_output = LOG.getvalue()
        self.assertFalse(log_output, "DEBUG-level log output should be empty.")

        # first-stage variables added by PyROS
        self.assertEqual(
            value(sep_model.first_stage.epigraph_var),
            value(nom_scenario_blk.first_stage.epigraph_var),
        )
        self.assertEqual(
            value(sep_model.first_stage.decision_rule_vars[0][0]),
            value(nom_scenario_blk.first_stage.decision_rule_vars[0][0]),
        )
        self.assertEqual(
            value(sep_model.first_stage.decision_rule_vars[0][1]),
            value(nom_scenario_blk.first_stage.decision_rule_vars[0][1]),
        )
        self.assertEqual(
            value(sep_model.first_stage.decision_rule_vars[0][2]),
            value(nom_scenario_blk.first_stage.decision_rule_vars[0][2]),
        )

        # variables of original user model
        self.assertEqual(value(sep_usr_blk.x1), value(nom_scenario_blk.user_model.x1))
        self.assertEqual(value(sep_usr_blk.x2), value(nom_scenario_blk.user_model.x2))
        self.assertEqual(value(sep_usr_blk.x3), value(nom_scenario_blk.user_model.x3))

        # uncertain parameter variable state should not have been
        # modified by the initialization
        self.assertEqual(
            value(sep_model.uncertainty.uncertain_param_indexed_var[0]),
            off_nominal_scenario[0],
        )
        self.assertEqual(
            value(sep_model.uncertainty.uncertain_param_indexed_var[1]),
            off_nominal_scenario[1],
        )
        self.assertTrue(sep_model.uncertainty.uncertain_param_var_list[0].fixed)
        self.assertTrue(sep_model.uncertainty.uncertain_param_var_list[1].fixed)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
from pyomo.common.dependencies import numpy as np, pandas as pd
import pyomo.contrib.parmest.parmest as parmest
from pyomo.contrib.parmest.examples.reactor_design.reactor_design import (
    reactor_design_model,
    ReactorDesignExperiment,
)

np.random.seed(1234)


class ReactorDesignExperimentDataRec(ReactorDesignExperiment):

    def __init__(self, data, data_std, experiment_number):

        super().__init__(data, experiment_number)
        self.data_std = data_std

    def create_model(self):

        self.model = m = reactor_design_model()
        m.caf.fixed = False

        return m

    def label_model(self):

        m = self.model

        # experiment outputs
        m.experiment_outputs = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.experiment_outputs.update(
            [
                (m.ca, self.data_i['ca']),
                (m.cb, self.data_i['cb']),
                (m.cc, self.data_i['cc']),
                (m.cd, self.data_i['cd']),
            ]
        )

        # experiment standard deviations
        m.experiment_outputs_std = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.experiment_outputs_std.update(
            [
                (m.ca, self.data_std['ca']),
                (m.cb, self.data_std['cb']),
                (m.cc, self.data_std['cc']),
                (m.cd, self.data_std['cd']),
            ]
        )

        # no unknowns (theta names)
        m.unknown_parameters = pyo.Suffix(direction=pyo.Suffix.LOCAL)

        return m


class ReactorDesignExperimentPostDataRec(ReactorDesignExperiment):

    def __init__(self, data, data_std, experiment_number):

        super().__init__(data, experiment_number)
        self.data_std = data_std

    def label_model(self):

        m = super().label_model()

        # add experiment standard deviations
        m.experiment_outputs_std = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.experiment_outputs_std.update(
            [
                (m.ca, self.data_std['ca']),
                (m.cb, self.data_std['cb']),
                (m.cc, self.data_std['cc']),
                (m.cd, self.data_std['cd']),
            ]
        )

        return m


def generate_data():

    ### Generate data based on real sv, caf, ca, cb, cc, and cd
    sv_real = 1.05
    caf_real = 10000
    ca_real = 3458.4
    cb_real = 1060.8
    cc_real = 1683.9
    cd_real = 1898.5

    data = pd.DataFrame()
    ndata = 200
    # Normal distribution, mean = 3400, std = 500
    data["ca"] = 500 * np.random.randn(ndata) + 3400
    # Random distribution between 500 and 1500
    data["cb"] = np.random.rand(ndata) * 1000 + 500
    # Lognormal distribution
    data["cc"] = np.random.lognormal(np.log(1600), 0.25, ndata)
    # Triangular distribution between 1000 and 2000
    data["cd"] = np.random.triangular(1000, 1800, 3000, size=ndata)

    data["sv"] = sv_real
    data["caf"] = caf_real

    return data


def main():

    # Generate data
    data = generate_data()
    data_std = data.std()

    # Create an experiment list
    exp_list = []
    for i in range(data.shape[0]):
        exp_list.append(ReactorDesignExperimentDataRec(data, data_std, i))

    # Define sum of squared error objective function for data rec
    def SSE_with_std(model):
        expr = sum(
            ((y - y_hat) / model.experiment_outputs_std[y]) ** 2
            for y, y_hat in model.experiment_outputs.items()
        )
        return expr

    ### Data reconciliation
    pest = parmest.Estimator(exp_list, obj_function=SSE_with_std)

    obj, theta, data_rec = pest.theta_est(return_values=["ca", "cb", "cc", "cd", "caf"])
    print(obj)
    print(theta)

    parmest.graphics.grouped_boxplot(
        data[["ca", "cb", "cc", "cd"]],
        data_rec[["ca", "cb", "cc", "cd"]],
        group_names=["Data", "Data Rec"],
    )

    ### Parameter estimation using reconciled data
    data_rec["sv"] = data["sv"]

    # make a new list of experiments using reconciled data
    exp_list = []
    for i in range(data_rec.shape[0]):
        exp_list.append(ReactorDesignExperimentPostDataRec(data_rec, data_std, i))

    pest = parmest.Estimator(exp_list, obj_function=SSE_with_std)
    obj, theta = pest.theta_est()
    print(obj)
    print(theta)

    theta_real = {"k1": 5.0 / 6.0, "k2": 5.0 / 3.0, "k3": 1.0 / 6000.0}
    print(theta_real)


if __name__ == "__main__":
    main()
